%FINDLOCALSHIFT estimates locally the shift 
%
% SYNOPSIS:
%  [ms, mc] = findlocalshift(in1, in2, boxsize, boxgrid, sigma, mask, CRlB)
%
% PARAMETERS:
%  in1:  2D input image 
%  in2:  2D input image 
%  boxsize: pixel size of the box between the shift is computed (must be odd) 
%  boxgrid: either a number e.g. 5 for a 5x5 grid points or a
%           matrix with the grid points in pixel [0,size-1]
%  sigma:   smoothing for the shift finding
%  mask:    mask image, shift is only computed if the box corners do not lie inside the mask 
%  CRlB:    compute Cramer-Rao lower bound for the std of the shift iter estimate as Pham2005 [yes/no]
%
% DEFAULTS:
%  boxsize: 10
%  boxgrid: 7
%  sigma:   1
%  mask:    []
%  CRlB:    0
%
% OUTPUT: 
%  ms: the shift vectors per grid box (pixel) [x: slice0, y: slice1 ]
%  mc: grid box center coordinates (pixel) [slice 0: x coord as with meshgrid]
%  mask_bgrid: the mask at the grid box center positions
%
% SEE ALSO:
%  localshift, orientation/vectorplot
%
% LITERATURE:
%  T.Q. Pham, M. Bezuijen, L.J. van Vliet, K. Schutte and C.L. Luengo Hendriks
%  Performance of optimal registration estimators
%  Z. Rahman, R.A. Schowengerdt and S.E. Reichenbach, editors. Volume 5817 of Proceedings of SPIE,
%  pp. 133-144, March 29-30 2005.


% (C) Copyright 2005               FEI Electron Optics - Building AAEp 
%     All rights reserved          PO Box 80066
%                                  5600 KA Eindhoven 
%                                  The Netherlands
% Bernd Rieger, May 2005

function varargout = findlocalshift(varargin)

d = struct('menu','Analysis',...
  'display','Find local shift',...
  'inparams',struct('name',       {'in1','in2','boxsize','boxgrid','sigma','mask','CRlB'},...
       'description',{'Image 1','Image 2','Box size [NxN] pixel','Box grid [NxN] points','Smoothing','Mask image','Compute CR lower bound'},...
       'type',       {'image','image','array','image','array','image','boolean'},...
       'dim_check',  {0,0,-1,0,-1,0,0},...
       'range_check',{'','','N+','','R+','',''},...
       'required',   {1,1,0,0,0,0,0},...
       'default',    {'a','b',31,'13',1,'[]',0}...
      ),...
  'outparams',struct('name',{'ms','mc_out','mask_mc'},...
                     'description',{'Local shift','Box center grid coordinates','mask_bgrid'},...
                     'type',{'image','image','image'}...
                     )...
 );
if nargin == 1
   s = varargin{1};
   if ischar(s) & strcmp(s,'DIP_GetParamList')
      varargout{1} = d;
      return
   end
end

try
   [in1, in2, boxsize, boxgrid, sigma, mask, CRlB] = getparams(d,varargin{:});
catch
   if ~isempty(paramerror)
      error(paramerror)
   else
      error(firsterr)
   end
end

sz1 = size(in1);
sz2 = size(in2);
if ndims(in1) ~=2 | ndims(in2) ~=2
   error('Images must be 2D.');
end
if sz1 ~= sz2
   error('Image are not same size.');
end


if ~mod(boxsize,2)
   warning('boxsize not odd, subtracting one.')
   boxsize = boxsize -1;
end
bx2 = (boxsize-1)/2;
%compute pixel coordinates of the box grid points
if ndims(boxgrid)==0
   %keep 10% safty margin to each borders
   perc = .8;
   boxgrid = double(boxgrid);
   nsz = perc * sz1;
   pc_d = nsz./boxgrid;
   [x,y]=meshgrid(1:boxgrid,1:boxgrid);
   pc_x = sz1(1)*(1-perc)/2 - pc_d(1)/2 + x*pc_d(1);
   pc_y = sz1(2)*(1-perc)/2 - pc_d(2)/2 + y*pc_d(2);
   tx = reshape(pc_x,1,prod(size(pc_x)));
   ty = reshape(pc_y,1,prod(size(pc_y)));
   mc = round(cat(1,tx,ty));
else
   sz = size(boxgrid);
   tx = reshape(double(ex_slice(boxgrid,0)),1,prod(sz(1:2)));
   ty = reshape(double(ex_slice(boxgrid,1)),1,prod(sz(1:2)));
   mc = round(cat(1,tx,ty));
   if size(mc,1) ~=2
      error('Box coordinates not in format [Nx2]')
   end
   boxgrid = sz(1);
end

if mc(1,1)-bx2 <0 | mc(end,1)+bx2 > sz1(1) mc(1,2)-bx2 <0 | mc(end,2)+bx2 > sz1(2)
   error('Box intersects image border, use smaller boxsize or less boxgridpoints.');
end
%sig = 1; %Gaussian noise estimate
sig = noisestd(in1);      

tbsx = 2*bx2 +1;
Nbox = size(mc,2);
ms1 = zeros([size(mc)]);
%ms = newim(size(mc),'dfloat');

ot1 = newim([boxsize boxsize Nbox]);
ot2 = newim([boxsize boxsize Nbox]);

mask_boxgrid = zeros(1,Nbox);
for ii = 1:Nbox 
   %test if the box corners are inside the provided mask
   if ~isempty(mask)
      if  mask(mc(1,ii)-bx2,mc(2,ii)-bx2) | mask(mc(1,ii)-bx2,mc(2,ii)+bx2) ...
        | mask(mc(1,ii)+bx2,mc(2,ii)-bx2) | mask(mc(1,ii)+bx2,mc(2,ii)+bx2)
        mask_boxgrid(ii) = 1; 
        continue;
      end
   end
   tmp1 = in1(mc(1,ii)-bx2 : mc(1,ii)+bx2, mc(2,ii)-bx2 : mc(2,ii)+bx2);
   tmp2 = in2(mc(1,ii)-bx2 : mc(1,ii)+bx2, mc(2,ii)-bx2 : mc(2,ii)+bx2);   
   s = findshift(tmp1,tmp2,'iter',sigma);
   %[v,p]=max(real(ift(ft(tmp1)*conj(ft(tmp2)))));
   %s = p - size(tmp1)./2;
   ms1(:,ii) = s;
   
   
   if CRlB
      ot1(:,:,ii-1) = tmp1;
      ot2(:,:,ii-1) = tmp2;
      %computing Cramer-Rao lower bound for the std of the shift iter estimate
      %as Pham2005
      g=gradient(tmp1);
      G=newimar(2,2);
      G{1,1} = sum(g{1}^2);
      G{2,2} = sum(g{2}^2);
      G{1,2} = sum(g{1}*g{2});
      G{2,1} = G{1,2};
      G = G./sig.^2;
      dG = det(G);
      sx = double(sqrt( G{2,2}/dG));
      sy = double(sqrt( G{1,1}/dG));
    
      fprintf('Processing box: %d, CRlb stdx: %f, stdy %f\n',ii,sx,sy);
   end
end
ms = newim(boxgrid, boxgrid,2,'dfloat');
mc_out = newim(boxgrid, boxgrid,2,'dfloat');

ms(:,:,0) = dip_image(reshape(ms1(1,:),boxgrid,boxgrid),'dfloat');
ms(:,:,1) = dip_image(reshape(ms1(2,:),boxgrid,boxgrid),'dfloat');

mc_out(:,:,0) = dip_image(reshape(mc(1,:),boxgrid,boxgrid),'dfloat');
mc_out(:,:,1) = dip_image(reshape(mc(2,:),boxgrid,boxgrid),'dfloat');

varargout{1} = ms;
if nargout >1
   varargout{2} = mc_out;
end
if nargout >2
   varargout{3} = dip_image(reshape(mask_boxgrid,[boxgrid boxgrid]),'bin');
end

